//---------------------------------------------------------------------------
// File: TMainForm.cpp, TDx_Input_Library v1.4, Example4 - Joystick
//
// Date: 17th May, 2001
// Author: Darren John Dwyer
// Copyright: (c) 2001 Darren John Dwyer, All Rights Reserved.
//
// Description:
//
// This example shows how to talk to the Joystick using DirectInput.
// It shows two/three (2/3) techniques of accessing the Joystick data :-
//
// (1) Polled Mode
// (2) Event-Driven Mode
// (3) Event-Drive + Forced Polled Mode
//
// -----------
// Polled Mode
// -----------
//
// To access the data from a DirectInput device, you need to ask it
// questions regarding it's current status.
// For example: is a particular button currently pressed on the Joystick?
//
// To this end, DirectInput requires you to force the update of the device
// status, periodically, using the TDx_InputDevice::Poll() method, to make
// certain that the device data retrieved using TDx_InputDevice::GetDeviceState()
// is accurate as possible.
//
// Hence, I've setup a TTimer() to poll the device and retrieve information.
//
// -----------------
// Event-Driven Mode
// -----------------
//
// Some devices can talk with the rest of the computer better than others.
// These devices are smart enough to generate standard win32-events into
// the windows event queue.
//
// The TDx_Library actively monitors and redirects these difficult to use
// events into easy to use TDx_InputDevice::OnDeviceStateChange() events,
// when enabled by calling TDx_InputDevice::SetEventNotifications(true).
//
// -------------------------------
// Event-Driven + Forced Poll Mode
// -------------------------------
//
// Overall, the event-driven method works best, and I suggest you use it first.
//
// Unfortunately, not all input devices (my mouse for example)
// generate the required win32 events, and hence some forced polling is required
// in order to accomodate these events.
//
// This example has been created to handle Event-Driven if possible, and use
// forced polling when not possible.
//
//---------------------------------------------------------------------------
#include <vcl\vcl.h>
//---------------------------------------------------------------------------
#pragma hdrstop
//---------------------------------------------------------------------------
#include "TMainForm.h"
//---------------------------------------------------------------------------
#if (__BORLANDC__ >= 0x0530) // BCB Version 3 +
    #pragma package(smart_init)
#endif
//---------------------------------------------------------------------------
#pragma link "TDx_Input"
#pragma link "TDx_InputDevice"
#pragma link "TDIDevCaps"
#pragma link "tdijoystate"
#pragma resource "*.dfm"
//---------------------------------------------------------------------------
#pragma link "TDx_Input_Library_Install"
//---------------------------------------------------------------------------
#define Msg(A)  {\
                if (StatusMemo->Text.Length()>30000) StatusMemo->Text = "";\
                StatusMemo->Lines->Add(A);\
                }
//---------------------------------------------------------------------------
TMainForm *MainForm;
//---------------------------------------------------------------------------
__fastcall TMainForm::TMainForm(TComponent* Owner)
    : TForm(Owner)
{
}
//---------------------------------------------------------------------------
__fastcall TMainForm::~TMainForm()
{
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::FormCreate(TObject *Sender)
{
Width=780;
Height=580;
Position=poScreenCenter;
GUID_Saved = false;
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::ExitButtonClick(TObject *Sender)
{
Close();
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::ClearMessagesButtonClick(TObject *Sender)
{
StatusMemo->Lines->Clear();
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::FormHide(TObject *Sender)
{
DestroyButtonClick(this);
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::CreateButtonClick(TObject *Sender)
{
// make sure to destroy any existing input devices first,
// anything that also needs to be deleted has been 'cascaded' nicely.
if (Dx_Input1->Created) Dx_Input1->Destroy();

// create a DirectInput interface
Dx_Input1->Create();

// disable the buttons so people cannot press the wrong ones.
CreateButton->Enabled = false;
DestroyButton->Enabled = true;
BeginPoll->Enabled = true;
EndPoll->Enabled = false;
BeginThread->Enabled = true;
EndThread->Enabled = false;
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::DestroyButtonClick(TObject *Sender)
{
// destroy any existing data, input devices, interfaces, etc.
// with the aim of 100% cleanup.
//
// anything that also needs to be deleted has been 'cascaded' nicely using
// the TDx_Input::OnCreate() and TDx_Input::OnDestroy() events.

if (Dx_Input1->Created) Dx_Input1->Destroy();

// disable the buttons so people cannot press the wrong ones.
CreateButton->Enabled = true;
DestroyButton->Enabled = false;
BeginPoll->Enabled = false;
EndPoll->Enabled = false;
BeginThread->Enabled = false;
EndThread->Enabled = false;
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::BeginPollClick(TObject *Sender)
{
Msg("Starting POLLED mode");

// disable the buttons so people cannot press the wrong ones.
BeginPoll->Enabled = false;
EndPoll->Enabled = true;
BeginThread->Enabled = false;
EndThread->Enabled = false;

// remember that we are in 'polled' mode.
Polled_Mode = true;

// enumerate devices and look for the first available joystick
Dx_Input1->EnumDevices( DIDEVTYPE_JOYSTICK, NULL, 0 );

// create an interface to the Joystick.
if (GUID_Saved) Dx_InputDevice1->Create( GUID_First_Joystick, Dx_Input1 );
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::EndPollClick(TObject *Sender)
{
Msg("Stopping POLLED mode");

// stop polling for data
PollingTimer->Enabled = false;

// destroy the joystick interface
if (Dx_InputDevice1->Created) Dx_InputDevice1->Destroy();

// re-enable buttons
BeginPoll->Enabled = true;
EndPoll->Enabled = false;
BeginThread->Enabled = true;
EndThread->Enabled = false;
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::BeginThreadClick(TObject *Sender)
{
Msg("Starting EVENT-DRIVEN mode");

// disable the buttons so people cannot press the wrong ones.
BeginPoll->Enabled = false;
EndPoll->Enabled = false;
BeginThread->Enabled = false;
EndThread->Enabled = true;

// remember that we are in 'event-driven' mode, not in 'polled' mode.
Polled_Mode = false;

// enumerate devices and look for the first available joystick
Dx_Input1->EnumDevices( DIDEVTYPE_JOYSTICK, NULL, 0 );

// create an interface to the Joystick.
if (GUID_Saved) Dx_InputDevice1->Create( GUID_First_Joystick, Dx_Input1 );
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::EndThreadClick(TObject *Sender)
{
Msg("Stopping EVENT-DRIVEN mode");

// disable the Forced Polling Timer
if (Forced_Polling_Required) ForcedPollTimer->Enabled = false;

// destroy the Joystick interface
if (Dx_InputDevice1->Created) Dx_InputDevice1->Destroy();

// re-enable buttons
BeginPoll->Enabled = true;
EndPoll->Enabled = false;
BeginThread->Enabled = true;
EndThread->Enabled = false;
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::Dx_Input1Error(TObject *Sender, AnsiString Function,
    AnsiString Error, AnsiString ErrorMessage)
{
Msg("ERROR ["+Error+"]: "+Function+", "+ErrorMessage);
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::Dx_Input1Create(TObject *Sender)
{
Msg("DirectInput Interface has been created");
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::Dx_Input1Destroy(TObject *Sender)
{
Msg("DirectInput interface is being destroyed." );

// auto-cascade destruction of related objects
if (Dx_InputDevice1->Created) Dx_InputDevice1->Destroy();
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::Dx_InputDevice1Create(TObject *Sender)
{
Msg("DirectInput Device has been Created");

// make the input device realize it's expecting Joystick data
Dx_InputDevice1->SetDataFormat( DIJoyState1 );

// make sure we have non-exclusive access to the Joystick
Dx_InputDevice1->SetCooperativeLevel( Application->Handle, DISCL_BACKGROUND | DISCL_NONEXCLUSIVE );

// depending on the mode chosen, setup for input device data retrieval
if (Polled_Mode)
    {
    // in "Polled Mode", we always poll for device input.
    Forced_Polling_Required = false;
    }
else
    {
    // setup automatic event notification redirection
    // if the device automatically generates events, SetEventNotification(true) returns true;
    // if the device does not generate events, SetEventNotifications(true) returns false;
    Forced_Polling_Required = Dx_InputDevice1->SetEventNotification(true);
    }

// acquire the Joystick enabling us to retrieve valid data from the Joystick
Device_Has_Been_Acquired = Dx_InputDevice1->Acquire();

// if acquired successfully, enable Input Data Polling
// NOTE: you can only enable polling after aquiring the device
if (Polled_Mode)
    {
    // turn on the polling mode
    PollingTimer->Enabled = Device_Has_Been_Acquired;
    }
else
    {
    // normally, no need to enable polling for event-driven results.
    // only enable polling if the device cannot automatically generate events.

    if (Forced_Polling_Required) ForcedPollTimer->Enabled = Device_Has_Been_Acquired;
    }
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::Dx_InputDevice1Destroy(TObject *Sender)
{
Msg("Input Device is being Destroyed");

// make sure to disable the Poll timer and the Forced Poll Timer
if (Polled_Mode)
    PollingTimer->Enabled = false;
else
    if (Forced_Polling_Required)
        ForcedPollTimer->Enabled = false;

// unacquire as required
if (Device_Has_Been_Acquired) Dx_InputDevice1->Unacquire();

// remember...
GUID_Saved = false;
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::PollingTimerTimer(TObject *Sender)
{
// this method is called only when using the "POLLED" option.
//
// normally, some input devices do not generate win32 queued events.
// in this case, you must call the TDx_InputDevice::Poll() method
// periodically in order to gain access to reliable data.
//
// an alternative, and much better, method used to access device data
// is available inside the TDx_Input_Library. This method is called
// "event-driven", and can be used when the device generates win32-events
// into the win32-event-queue. Inside the TDx_Input_Library, these
// win32-events have been converted to a BCB OnDeviceStateChange() event,
// much easier to use, although sometimes you still need to "poll".
//
// in this example, i've shown how to create a standard "poll mode"
// routines and also how to create "event-driven / forced poll mode" routine.
//

AnsiString str = "POLLED: ";

// check to see if the Joystick device state has changed
if (Dx_InputDevice1->GetDeviceState( DIJoyState1 ))
    {
    // it has changed, and the DIJoyState1 has been updated.
    // so, do something with the data (display it)

    str = str + "[X: "+IntToStr(DIJoyState1->X)+", Y: "+IntToStr(DIJoyState1->Y)+", Z: "+IntToStr(DIJoyState1->Z)+"]";

    // this little macro can be used to determine if a button is pressed
    #define PRESSED(A)   ((A) & 0x80)

    if (PRESSED(DIJoyState1->Buttons[0])) str = str + " 0 ";
    if (PRESSED(DIJoyState1->Buttons[1])) str = str + " 1 ";
    if (PRESSED(DIJoyState1->Buttons[2])) str = str + " 2 ";
    if (PRESSED(DIJoyState1->Buttons[3])) str = str + " 3 ";

    ButtonsPressed->Caption = str;
    ButtonsPressed->Update();
    }
else
    {
    ButtonsPressed->Caption = "ERROR";
    ButtonsPressed->Update();
    }
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::Dx_InputDevice1DeviceStateChange(TObject *Sender)
{
// this method is called automatically when using the "EVENT DRIVEN" option.
//
// because some input devices do not generate win32 queued events, sometimes
// you must call the TDx_InputDevice::Poll() method, which causes the device to
// generate the relevant win32 events.
//
// Forced Poll events are also redirected automatically to this event.
//

AnsiString str = "EVENT-DRIVEN: ";

// the code below is identical to the polled version
if (Dx_InputDevice1->GetDeviceState( DIJoyState1 ))
    {
    str = str + "[X: "+IntToStr(DIJoyState1->X)+", Y: "+IntToStr(DIJoyState1->Y)+", Z: "+IntToStr(DIJoyState1->Z)+"]";

    #define PRESSED(A)   ((A) & 0x80)

    if (PRESSED(DIJoyState1->Buttons[0])) str = str + " 0 ";
    if (PRESSED(DIJoyState1->Buttons[1])) str = str + " 1 ";
    if (PRESSED(DIJoyState1->Buttons[2])) str = str + " 2 ";
    if (PRESSED(DIJoyState1->Buttons[3])) str = str + " 3 ";

    ButtonsPressed->Caption = str;
    ButtonsPressed->Update();
    }
else
    {
    ButtonsPressed->Caption = "ERROR";
    ButtonsPressed->Update();
    }
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::ForcedPollTimerTimer(TObject *Sender)
{
Dx_InputDevice1->Poll();
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::Dx_Input1EnumDevices(TObject *Sender, TDIDeviceInstance *DeviceInstance, void *Context,
    bool &Finished)
{
if ((DeviceInstance->DevType & DIDEVTYPE_JOYSTICK)==DIDEVTYPE_JOYSTICK)
    {
    GUID_Saved = true;
    GUID_First_Joystick = DeviceInstance->Product;
    }
else
    Finished = true;
}
//---------------------------------------------------------------------------
